local TOC, CCD = ...

CCD.Module = CCD.Modules or {}
CCD.Modules.RBT = CCD.Modules.RBT or {}
local RBT = CCD.Modules.RBT
CCD.Buff = CCD.Buff or {}
CCD.Allies = CCD.Allies or {}
local Allies = CCD.Allies
CCD.Target = CCD.Target or {}
CCD.Debug = CCD.Debug or {}
local Debug = CCD.Debug

CCD.Ally = CCD.Ally or {}
function CCD.Ally:Create(Id)
	local ally = {}
	setmetatable(ally, self)
	self.__index = self

	ally.Buffs = {}
	ally.BuffsByType = {}

	local details = Inspect.Unit.Detail(Id)
	ally.Availability = details.availability
	ally.IsBlocked = details.blocked
	ally.Class = details.calling
	ally.Health = details.health
	ally.Id = details.id
	ally.FullName = details.name -- used in Addon Messages
	local index = string.find(details.name, "@")
	if index then
		ally.Name = string.sub(details.name, 1, index - 1)
	else
		ally.Name = details.name
	end
	ally.IsOffline = details.offline
	ally.Role = details.role
	ally.InCombat = details.combat
	if ally.InCombat == nil then ally.InCombat = false end
	ally.IsArchon = false -- because archon detection is complicated
	ally.CanProvideLivingEnergy = false -- because detecting if a toon can provide Living Energy is complicated

	return ally
end
function CCD.Ally:ReadDetails()
	local details = Inspect.Unit.Detail(self.Id)
	self.Availability = details.availability
	self.IsBlocked = details.blocked
	self.Class = details.calling
	self.Health = details.health
	self.Id = details.id
	self.FullName = details.name -- used in Addon Messages
	local index = string.find(details.name, "@")
	if index then
		self.Name = string.sub(details.name, 1, index - 1)
	else
		self.Name = details.name
	end
	self.IsOffline = details.offline
	self.Role = details.role
	self.InCombat = details.combat
	if self.InCombat == nil then self.InCombat = false end
	if details.zone then
		local zoneDetails = Inspect.Zone.Detail(details.zone)
		if zoneDetails then
			self.ZoneType = zoneDetails.type or "open"
			--if self == CCD.Me then print("New Zone Type: " .. self.ZoneType) end
		end
	end

	if self.OnReadDetailsEnd then self:OnReadDetailsEnd() end
end
function CCD.Ally:GetResource() -- MANA, ENERGY or UNKNOWN
	if self.Class == "mage" or self.Class == "cleric" then
		return "MANA"
	elseif self.Class == "rogue" or self.Class == "warrior" then
		return "ENERGY"
	else
		return "UNKNOWN"
	end
end
function CCD.Ally:AddBuff(Buff)
	if not Buff then -- should never happen
		if Debug.DeveloperModeOn then
			Debug:Record("CCD.Ally:AddBuff(Buff); Buff = nil")
			print("Error recorded.")
		end
		return
	end

	if Buff.Type then -- Some buffs don't have types e.g. Rested Exp and Track Ore
		self.BuffsByType[Buff.Type] = Buff
		self.Buffs[Buff.Id] = Buff
	end

	Debug:Record(string.format("%s has gained %s (Id: %s; Type: %s)", self.Name, Buff.Name, Buff.Id, Buff.Type))
end
function CCD.Ally:RemoveBuff(BuffId)
	local buff = self.Buffs[BuffId]

	if not buff then return false end -- irrelevant buff

	
	self.Buffs[buff.Id] = nil
	self.BuffsByType[buff.Type] = nil -- currently, only buffs with types are added; if buff.Type then self.BuffsByType[buff.Type] = nil end

	-- check if a buff with the removed type (rune) still exists
	for _, buffV in pairs(self.Buffs) do
		if buffV.Type == buff.Type then
			self.BuffsByType[buffV.Type] = buffV
		end
	end
	
	buff:Delete()
	
	if CCD.Debug.DeveloperModeOn then
		Debug:Record(string.format("%s has lost %s (Id: %s; Type: %s)", self.Name, buff.Name, buff.Id, buff.Type))
	end
	
	return buff
end
function CCD.Ally:ClearAllBuffs()
	for id, buff in pairs(self.Buffs) do
		buff:Delete()
		self.Buffs[id] = nil
	end

	for buffType in pairs(self.BuffsByType) do
		self.BuffsByType[buffType] = nil
	end

	if Debug.DeveloperModeOn then
		Debug:Record(string.format("%s's buffs have been cleared", self.Name))
	end
end
function CCD.Ally:InitializeBuffs() -- only used on CCD.Me atm
	corout.new(self.InitializeBuffs_Task, "Ally:InitializeBuffs_Task()", self)
end
function CCD.Ally:InitializeBuffs_Task()
	local buffIds = Inspect.Buff.List(self.Id)

	if not buffIds then -- this should never happen
		if Debug.DeveloperModeOn then
			Debug:Record("Ally:InitializeBuffs_Task() - local buffIds = nil") 
			print("Error recorded.")
			return
		end
	end

	for buffId in pairs(buffIds) do
		local buff = CCD.Buff:Create(self.Id, buffId)

		-- check for raid buff/debuff
		if RBT.Buffs[buff.Type] or RBT.Debuffs[buff.Type] ~= nil then
			self:AddBuff(buff)
		end

		corout.check()
	end
end

function CCD.Ally:GetRole()
	return self.Role
end
function CCD.Ally:IsFull()
	return self.Availability == "full"
end
function CCD.Ally:IsPartial()
	return self.Availability == "partial"
end
function CCD.Ally:IsNone()
	return self.Availability == "none"
end