local TOC, CCD = ...

-- Globals
CCD.Messaging = CCD.Messaging or {}
CCD.Messaging.Categories = {
	CooldownRequest = "CooldownRequest",
	CooldownList = "CooldownList",
	SingleCooldown = "SingleCooldown",
	FullyAvailable = "FullyAvailable",
	CdBegin = "CooldownBegin",
	CdEnd = "CooldownEnd",
	Archon = "iArchon",
	RoleChanged = "RoleChanged",
	Search = "Search",
	Found = "Found",
}
CCD.Messaging.Clients = CCD.Messaging.Clients or {} -- Names by Id
CCD.Messaging.ClientCount = 0
CCD.Messaging.WaitingList = CCD.Messaging.WaitingList or {}
CCD.Messaging.ShyModeOn = false
CCD.Messaging.Enabled = true -- effectively turns the Messaging system on or off; to be removed
CCD.Modules.RBT = CCD.Modules.RBT or {}
CCD.Modules.RCT = CCD.Modules.RCT or {}
CCD.UnitAvailability = {
	None = "none",
	Partial = "partial",
	Full = "full",
}

-- Locals
local CooldownInfoDNSList = {} -- List of clients that a full list of cooldown info should NOT be sent to; Indexed by ClientId
local CooldownSendList = {}
local BroadcastCooldownListWhenFull = true

-- Aliases
local Categories = CCD.Messaging.Categories
local Clients = CCD.Messaging.Clients
local WaitingList = CCD.Messaging.WaitingList
local RBT = CCD.Modules.RBT
local RCT = CCD.Modules.RCT
local Allies = CCD.Allies.ValuesById
local UAV = CCD.UnitAvailability

-- Messages
CCD.Messaging.CooldownListMessage = {
	CooldownTable, -- ability time remaining by ability ID
	SerializeString,
}
function CCD.Messaging.CooldownListMessage:Create()
	local message = {}
	setmetatable(message, self)
	self.__index = self
	
	self.CooldownTable = RCT:GetAllCooldownInfo()
	local cooldownStrings = {}
	
	for id, remaining in pairs(self.CooldownTable) do
		table.insert(cooldownStrings, string.format("%s:%s", id, remaining))
	end
	
	self.SerializedString = table.concat(cooldownStrings, ";")
	
	return self
end
function CCD.Messaging.CooldownListMessage:Deserialize(SerializedString)
	local message = {}
	setmetatable(message, self)
	self.__index = self
	
	self.SerializedString = SerializedString
	self.CooldownTable = {}
	local cooldownStrings = CCD.Messaging.SplitData(SerializedString, ";")

	for _, cdString in ipairs(cooldownStrings) do
		local splits = CCD.Messaging.SplitData(cdString, ":")
		local id = splits[1]
		local remaining = tonumber(splits[2])
		self.CooldownTable[id] = remaining
	end
	
	return self
end
--
CCD.Messaging.SingleCooldownMessage = {
	AbilityId,
	AbilityCooldownRemaining,
	SerializeString,
}
function CCD.Messaging.SingleCooldownMessage:Create(AbilityId, AbilityCooldownRemaining)
	local message = {}
	setmetatable(message, self)
	self.__index = self
	
	self.AbilityId = AbilityId
	self.AbilityCooldownRemaining = AbilityCooldownRemaining
	self.SerializedString = string.format("%s:%s", self.AbilityId, self.AbilityCooldownRemaining)
	
	return self
end
function CCD.Messaging.SingleCooldownMessage:Deserialize(SerializedString)
	local message = {}
	setmetatable(message, self)
	self.__index = self
	
	self.SerializedString = SerializedString
	
	local splits = CCD.Messaging.SplitData(self.SerializedString, ":")
	self.AbilityId = splits[1]
	self.AbilityCooldownRemaining = tonumber(splits[2])
	
	return self
end

-- Functions
function CCD.Messaging.AddClient(ClientId, ClientName)
	Clients[ClientId] = ClientName
	CCD.Messaging.ClientCount = CCD.Messaging.ClientCount + 1
end
function CCD.Messaging.RemoveClient(UnitId)
	if Clients[UnitId] then
		Clients[UnitId] = nil
		CCD.Messaging.ClientCount = CCD.Messaging.ClientCount - 1
	end
end
function CCD.Messaging.SplitData(Data, Separator)
	Separator = Separator or "%s"
	
	local t = {}
	local i = 1
	
	for str in string.gmatch(Data, "([^" .. Separator .. "]+)") do
		t[i] = str
		i = i + 1
	end
	
	return t
end

-- Sending
function CCD.Messaging.Callback(Failure, Message)
	if Failure then
		CCD.Debug:Record("Callback failed!")
		--print("Callback failed!")
	end
end
function CCD.Messaging.BroadcastToRaid(Category, Message)
	if CCD.Allies.Count == 1 then
		return
	elseif CCD.Allies.Count <= 5 then
		msgType = "party"
	else
		msgType = "raid"
	end
	
	local actualMessage = string.format("%s,%s,%s", Category, CCD.Me.Id, Message)
	--print("Size = " .. Utility.Message.Size(TargetName, TOC.identifier, actualMessage))
	Command.Message.Broadcast(msgType, nil, TOC.identifier, actualMessage, CCD.Messaging.Callback)
end

function CCD.Messaging.BroadcastCooldownListRequest() -- Request cooldowns from everyone in group
	CCD.Messaging.BroadcastToRaid(Categories.CooldownRequest, "")
end
function CCD.Messaging.BroadcastCooldownList()
	if not CCD.Me:IsFull() then	return end
	
	local clm = CCD.Messaging.CooldownListMessage:Create()
	if clm.SerializedString == "" then return end
	CCD.Messaging.BroadcastToRaid(Categories.CooldownList, clm.SerializedString)
	CCD.Messaging.BroadcastFullAvailability()
end
function CCD.Messaging.BroadcastSingleCooldown(AbilityId, AbilityCooldownRemaining)
	local singleCdMsg = CCD.Messaging.SingleCooldownMessage:Create(AbilityId, AbilityCooldownRemaining)
	CCD.Messaging.BroadcastToRaid(Categories.SingleCooldown, singleCdMsg.SerializedString)
end
function CCD.Messaging.BroadcastAbilityBegin(AbilityId, AbilityCooldownRemaining)
	CCD.Messaging.BroadcastToRaid(Categories.CdBegin, string.format("%s:%s", AbilityId, AbilityCooldownRemaining)) -- only broadcasted to clients in the same world/instance
end

function CCD.Messaging.BroadcastFullAvailability()
	if not CCD.Me:IsFull() then return end
	CCD.Messaging.BroadcastToRaid(Categories.FullyAvailable, "")	
end
function CCD.Messaging.BroadcastRoleChanged()
	CCD.Messaging.BroadcastToRaid(Categories.RoleChanged, "")
end
function CCD.Messaging.BroadcastArchonStatus()
	if CCD.Me.IsArchon then
		CCD.Messaging.BroadcastToRaid(Categories.Archon, "")
	end
end

function CCD.Messaging.FindClients()
	CCD.Messaging.Searching = true
	CCD.Messaging.BroadcastToRaid(Categories.Search, "")
	Library.LibCron.new(TOC.identifier, 2, 1, false, function() CCD.Messaging.Searching = false end)
end
function CCD.Messaging.BroadcastFound()
	CCD.Messaging.BroadcastToRaid(Categories.Found, CCD.Version)
end

-- Receiving
function CCD.Messaging.OnMessageReceived(Handle, Sender, MsgType, Channel, Id, Data)
	if Id ~= TOC.identifier then return end
	if Sender == CCD.Me.Name then return end
	local splits = CCD.Messaging.SplitData(Data, ",")
	local category = splits[1]
	local clientId = splits[2]
	
	if not clientId then return end
	local ally = Allies[clientId]
	if not ally then return end
	
	--print("Received message of type " .. splits[1] .. " from " .. Sender)
		
	if		category == Categories.CooldownRequest then CCD.Messaging.OnCooldownListRequest()
	elseif	category == Categories.CooldownList then CCD.Messaging.OnCooldownList(clientId, splits[3])
	elseif	category == Categories.SingleCooldown then CCD.Messaging.OnSingleCooldown(clientId, splits[3])
	elseif	category == Categories.CdBegin then	CCD.Messaging.OnCooldownBegin(clientId, splits[3])
	elseif	category == Categories.FullyAvailable then CCD.Messaging.OnFullyAvailable(ally)
	elseif	category == Categories.RoleChanged then CCD.Messaging.OnRoleChanged(ally)
	elseif	category == Categories.Archon then CCD.Messaging.OnArchon(ally)
	elseif	category == Categories.Search then CCD.Messaging.BroadcastFound()
	elseif	category == Categories.Found then CCD.Messaging.OnFound(clientId, Sender, splits[3])
	end
end

function CCD.Messaging.OnCooldownListRequest()
	CCD.Messaging.BroadcastCooldownList()	
	CCD.Messaging.BroadcastArchonStatus()
end
function CCD.Messaging.OnCooldownList(ClientId, CooldownListString)
	local cooldownListMessage = CCD.Messaging.CooldownListMessage:Deserialize(CooldownListString)

	for id, remaining in pairs(cooldownListMessage.CooldownTable) do
		RCT:OnAbilityAdded(id, remaining, ClientId)
	end
end
function CCD.Messaging.OnSingleCooldown(ClientId, SingleCooldownString)	
	local singleCooldownMessage = CCD.Messaging.SingleCooldownMessage:Deserialize(SingleCooldownString)

	RCT:OnAbilityAdded(singleCooldownMessage.AbilityId, singleCooldownMessage.AbilityCooldownRemaining, ClientId)
end
function CCD.Messaging.OnCooldownBegin(ClientId, CooldownBeginString)	
	local splits = CCD.Messaging.SplitData(CooldownBeginString, ":")
	local abilityId = splits[1]
	local abilityCooldownRemaining = tonumber(splits[2])
	
	RCT:OnAbilityNewCooldownBegin(ClientId, abilityId, abilityCooldownRemaining)
end

function CCD.Messaging.OnFullyAvailable(Ally)
	local partialTimerTag = CCD.PartialAvailabilityTimersByAlly[Ally]
			
	if partialTimerTag then
		Library.LibCron.delete(partialTimerTag)				
		CCD.PartialAvailabilityTimersByAlly[Ally] = nil
	end	
end
function CCD.Messaging.OnRoleChanged(Ally)
	RBT:OnRoleCastCompleted(Ally)
	RCT:OnRoleCastCompleted(Ally)
end
function CCD.Messaging.OnArchon(Ally)	
	RBT:AddArchon(Ally)
end

function CCD.Messaging.OnFound(ClientId, ClientName, Version)
	CCD.Messaging.AddClient(ClientId, Version)
end

-- Event Handlers
function CCD.Messaging.OnJoinedGroup()
	CCD.Messaging.BroadcastFound()
	CCD.Messaging.BroadcastCooldownList()
	CCD.Messaging.BroadcastCooldownListRequest()
	CCD.Messaging.BroadcastArchonStatus()
	CCD.Messaging.FindClients()
end
function CCD.Messaging.OnLeftGroup(UnitId)	
	CCD.Messaging.RemoveClient(UnitId)
end
function CCD.Messaging.OnAllyFullAvailability(Ally)	
	if Clients[Ally.Id] then
		--CCD.Messaging.RequestCooldowns(Ally)		
	else
		
	end	
end
function CCD.Messaging.OnPlayerFullAvailability()
	CCD.Messaging.FullAvailabilityTimeStamp = Inspect.Time.Frame()	
end
function CCD.Messaging.OnUpdate()
	if CCD.Messaging.FullAvailabilityTimeStamp and Inspect.Time.Frame() > CCD.Messaging.FullAvailabilityTimeStamp then
		CCD.Messaging.BroadcastCooldownList()
		CCD.Messaging.FullAvailabilityTimeStamp = nil
	end
end

-- Events and Initialisation
Command.Event.Attach(Event.Message.Receive, CCD.Messaging.OnMessageReceived, "CCD.Messaging.OnMessageReceived")
Command.Message.Accept(nil, TOC.identifier)

--[[ Notes
Clients that are partially available to each other cannot communicate privately until they've sent a tell.

Cooldown Lists are broadcasted when:
1. On first frame update after Player enters Full availability.
2. Player reads a Cooldown List Request.
3. Player enters a group.

When player loads a map:
00 Player enters Partial availability. Group members start a 5 second timer.
01 Player is still loading...
02 Player is still loading...
03 Player is still loading...
04 Player enters Full availability and broadcasts a Full availability message. Group members read this broadcast and terminate their timers.

Archon status is broadcasted when:
1. Player joins a group.
2. Player reads a Cooldown List request.

Bandwidth Limits
Burst = 10,240		The amount of data that can be sent or received without being subject to throttling.
Maximum = 8,192		Maximum allowed message size.
Sustained = 2,048	Amount of data that can be transferred per second without dropping messages.
--]]